/*

 CHASM C/P - CHASM Command Parser
 Copyright (c) 2004 PXI Project Team

 Originally written in RealBasic and ported to C++ by Jonathan Bettencourt.

 This parses a string of xTalk code according to command definitions in what's called
 a "vocabulary file," whose format was written by Jonathan Bettencourt.

 To turn a line of xTalk into a vector, either:
 * Call ccp_init(), then pass a script line to ccp_parse(<script line>, <parsing mode>).
 * Create an instance of chasm_vocabulary, load a vocabulary file with the readfile(<file name>)
   method, and parse with chasm_parse(<script line>, <parsing mode>).

 The parsing mode is an optional parameter. There are three possible values:

     0 (or the constant CCP_PARSE_NUMERIC) - Puts parameters in a numbered order.
     If a parameter has no number, it appears after the numbered parameters as a set of two
     strings, one with a colon and the parameter name, and the second with the value.

     1 (or the constant CCP_PARSE_NOMINATIVE) - Assigns names to all parameters,
     regardless of whether they have numbers or not. Each parameter is defined by a set of
     two strings, one with a colon and the parameter name, and the second with the value.

     2 (or the constant CCP_PARSE_DUAL) - Combines the first two modes. All numbered
     parameters will be put in order, followed by all parameters with their names,
     regardless of whether they have numbers or not.

 To turn debugging messages on, replace    / / DEBUGGING_CODE! : :   with   / * DEBUGGING_CODE! * /
 To turn debugging messages off, replace   / * DEBUGGING_CODE! * /   with   / / DEBUGGING_CODE! : :

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in the
 Software without restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#include <string>
#include <vector>
#include <map>
#include <stack>
#include <iostream>
#include "stringmaniplib.h"
#include "binarystream.h"
#include "bchasm.h"
#include "variant.h"
#include "chasmcp.h"
using namespace std;

chasm_vocabulary ccp_global_vocab;

vector<string> ccp_jrrTolkienList;

/* CHASM Vocabulary Word
   This is a class that holds all the information about how
   to parse a certain command. */

string chasm_vocab_word::gparams(int i)
{
	if (i>=0 && i<(signed long)params.size()) { return params[i]; } else { return ""; }
}

void chasm_vocab_word::sparams(int i, string s)
{
	if (i>=0 && i<(signed long)params.size()) { params[i] = s; }
}

string chasm_vocab_word::htequiv(void)
{
	return gparams(params.size() - 1);
}

string chasm_vocab_word::kind(void)
{
	string s;
	if (type == "command") { return (string)"SphereTalk command"; }
	if (type == "function") { return (string)"SphereTalk function"; }
	if (type == "setter") { return (string)"SphereTalk property setter"; }
	if (type == "constant") { return (string)"constant"; }
	if (type == "ccommand") { return (string)"CHASM command"; }
	if (type == "rbscript") { return (string)"RBScript command preprocessor"; }
	if (type == "chasm") { return (string)"CHASM command preprocessor"; }
	s = "vocabulary item [";
	s += type + "]";
	return s;
}

int chasm_vocab_word::sortpriority(void)
{
	if (type == "chasm") { return 1; }
	if (type == "rbscript") { return 2; }
	if (type == "command") { return 3; }
	if (type == "setter") { return 4; }
	if (type == "function") { return 5; }
	if (type == "constant") { return 6; }
	if (type == "ccommand") { return 7; }
	return 0;
}

void chasm_vocab_word::readdata(const string & s)
{
	int i,m;
	string d;
	string delim = " ";
	delim[0] = 2;
	paramcount = 0;
	params.clear();
	type = trim( nthfield(s, delim, 1) );
	name = trim( nthfield(s, delim, 2) );
	m = countfields(s, delim) - 2;
	if (m > 0)
	{
		for (i=0; i<m; ++i)
		{
			paramcount++;
			d = nthfield(s, delim, i+3);
			if (d != " ")
			{
				params.push_back(trim(d));
			} else {
				params.push_back(d);
			}
		}
	}
}

string chasm_vocab_word::writedata(void)
{
	string s, d;
	int i, m;
	string delim = " ";
	delim[0] = 2;
	s = trim(type) + delim + trim(name);
	m = paramcount;
	if (m > 0)
	{
		for (i=0; i<m; ++i)
		{
			s += delim;
			d = gparams(i);
			if (d != " ")
			{
				s += trim(d);
			} else {
				s += d;
			}
		}
	}
	return s;
}

/* CHASM Vocabulary
   This is the class that actually parses commands, using
   vocabulary defined by chasm_vocab_words. */

chasm_vocabulary::chasm_vocabulary(void)
{
	/* if you feel like changing the default settings, here they are */
	convertoptionl = false;
	stripSpaces = true;
	processAssignment = false;
	processFunctiOn = true;
}

chasm_vocabulary::chasm_vocabulary(const string & vocabfile)
{
	/* if you feel like changing the default settings, change them here as well */
	convertoptionl = false;
	stripSpaces = true;
	processAssignment = false;
	processFunctiOn = true;

	readfile(vocabfile);
	sort();
}

int chasm_vocabulary::countchars(const string & s, const string & whatToCnt)
{
	int c = 0;
	unsigned long int i;
	if (s == "")
	{
		return 0;
	} else {
		for (i=0; i<s.size(); ++i)
		{
			if (s[i] == whatToCnt[0]) { c++; } //heh...
		}
	}
	return c;
}

bool chasm_vocabulary::equalparens(const string & s)
{
	// return ( countchars(s, "(") == countchars(s, ")") );
	int i, lpc, rpc;
	string l;
	lpc = 0;
	rpc = 0;
	for (i = 0; i < (signed int)s.size(); i++)
	{
		l = left(s, i);
		if (equalquotes(l))
		{
			if (s[i] == 28) { lpc++; }
			if (s[i] == 29) { rpc++; }
		}
	}
	return ( lpc == rpc );
}

bool chasm_vocabulary::equalquotes(const string & s)
{
	string quot = " ";
	quot[0] = 34;
	return (( countchars(s, quot) % 2 ) == 0);
}

bool chasm_vocabulary::balancedexpr(const string & s)
{
	// DEBUGGING_CODE! :: cout << "Input to BExpr: " << s << endl;
	string incrementorWords = " char character word line item para paragraph byte short long sent sentence list elem element ";
	string decrementorWords = " of in ";
	int i, m;
	string cw;
	int balance = 0;
	bool lastWordWasThe = false;
	m = countfields(s, " ");
	for (i=1; i<=m; i++)
	{
		cw = nthfield(s, " ", i);
		if ( cw[0] == '(' ) { cw = mid(cw, 2); }
		cw = (string)" " + cw + (string)" ";
		if ( (!lastWordWasThe) || ((cw != (string)" short ") && (cw != (string)" long ")) )
		{
			/* For flag-based check, use the following: */
			if (instr(incrementorWords, cw) > 0) { balance = 1; }
			if (instr(decrementorWords, cw) > 0) { balance = 0; }

			/* For balance-based check, use the following: */
			// if (instr(incrementorWords, cw) > 0) { balance++; }
			// if (instr(decrementorWords, cw) > 0) { balance--; }
			// if (balance < 0) { balance = 0; }
		}
		if (cw == (string)" the ")
		{
			lastWordWasThe = true;
		} else {
			if (cw != (string)"  ") { lastWordWasThe = false; }
		}
	}
	// DEBUGGING_CODE! :: cout << "Balance: " << balance << endl;
	return (balance == 0);
}

string chasm_vocabulary::stripdupspaces(const string & theString)
{
	string s = theString;
	int i;
	string l,r;
	for (i = s.size()-1; i>0; i--)
	{
		if ( mid(s, i, 2) == (string)"  " )
		{
			l = left(s, i-1);
			r = mid(s, i+1);
			if (equalquotes(l)) { s = l+r; }
		}
	}
	return trim(s);
}

string chasm_vocabulary::findFactor(const string & s)
{
	vector<string> v;
	int w;
	int i;
	string s1, s2;
	string d = " ";
	d[0] = 127;
	v = ccp_jrrTolkienizer(s);
	w = findFactor(v, 0);
	for (i=0; i<w; i++)
	{
		s1 += (string)" ";
		s1 += mid( (string)v[i] ,3);
	}
	for (i=w; i<(signed int)v.size(); i++)
	{
		s2 += (string)" ";
		s2 += mid( (string)v[i] ,3);
	}
	return (s1 + d + s2);
}

string chasm_vocabulary::findFactor(const string & s, const string & ss)
{
	vector<string> v;
	vector<string> v1;
	int w;
	int i;
	string s1, s2;
	string d = " ";
	d[0] = 127;
	v = ccp_jrrTolkienizer(s + (string)" " + ss);
	v1 = ccp_jrrTolkienizer(s);
	w = findFactor(v, v1.size());
	for (i=v1.size(); i<w; i++)
	{
		s1 += (string)" ";
		s1 += mid( (string)v[i] ,3);
	}
	for (i=w; i<(signed int)v.size(); i++)
	{
		s2 += (string)" ";
		s2 += mid( (string)v[i] ,3);
	}
	return (s1 + d + s2);
}

int chasm_vocabulary::findFactor(const vector<string> & v, int i)
{
	/* here our return value is one token after the last token in the factor */
	// DEBUGGING_CODE! :: cout << v[i] << endl;
	int b;
	if ( i >= (signed int)v.size() ) { return i; }
	while ( left( (string)v[i], 2 ) == "SK" ) { i++; }
	if ( i >= (signed int)v.size() ) { return i; }
	if ( left( (string)v[i], 2 ) == "QU" ) { return (i+1); }
	if ( (string)v[i] == (string)"OS(" )
	{
		b = 0;
		while ( ((string)v[i] != (string)"OS)") && (b <= 0) )
		{
			i++;
			if ( i >= (signed int)v.size() ) { return i; }
			if ( (string)v[i] == (string)"OS(" ) { b++; }
			if ( (string)v[i] == (string)"OS)" ) { b--; }
		}
		return (i+1);
	}
	if ( i >= (signed int)v.size() ) { return i; }
	if ( left( (string)v[i], 2 ) == "CH" )
	{
		// DEBUGGING_CODE! :: cout << "1: " << v[i] << endl;
		i++;
		if ( i >= (signed int)v.size() ) { return i; }
		// DEBUGGING_CODE! :: cout << "2: " << v[i] << endl;
		if ( instr( (string)" OP OS EW " , left( (string)v[i], 2 ) ) > 0 )
		{
			i++;
			if ( i >= (signed int)v.size() ) { return i; }
			if (left( (string)v[i], 2 ) == "CW") { i++; return findFactor(v, i); }
			else { return i; }
		}
		// DEBUGGING_CODE! :: cout << "3: " << v[i] << endl;
		b = 0;
		while ( (left( (string)v[i], 2 ) != "EW") && (left( (string)v[i], 2 ) != "CW") && (b <= 0) )
		{
			i++;
			if ( i >= (signed int)v.size() ) { return i; }
			if ( (string)v[i] == (string)"OS(" ) { b++; }
			if ( (string)v[i] == (string)"OS)" ) { b--; }
		}
		// DEBUGGING_CODE! :: cout << "4: " << v[i] << endl;
		if ( (string)v[i] == (string)"EWto" )
		{
			b = 0;
			i++;
			if ( i >= (signed int)v.size() ) { return i; }
			while ( (left( (string)v[i], 2 ) != "EW") && (left( (string)v[i], 2 ) != "CW") && (b <= 0) )
			{
				i++;
				if ( i >= (signed int)v.size() ) { return i; }
				if ( (string)v[i] == (string)"OS(" ) { b++; }
				if ( (string)v[i] == (string)"OS)" ) { b--; }
			}
		}
		// DEBUGGING_CODE! :: cout << "5: " << v[i] << endl;
		//i++;
		//if ( i >= (signed int)v.size() ) { return i; }
		if (left( (string)v[i], 2 ) == "CW") { i++; return findFactor(v, i); }
		else { return i+1; }
	}

	if ( left( (string)v[i], 2 ) == "OB" )
	{
		while ( left( (string)v[i], 2 ) == "OB" )
		{
			i++;
			if ( i >= (signed int)v.size() ) { return i; }
			if ( (left( (string)v[i], 2 ) == "OP") || (left( (string)v[i], 2 ) == "OS") || (left( (string)v[i], 2 ) == "EW") ) { return i; }
		}
		i++;
		if ( i >= (signed int)v.size() ) { return i; }
		if ( (string)v[i] == (string)"OS(" ) { return findFactor(v, i); }
		if ( (left( (string)v[i], 2 ) == "OP") || (left( (string)v[i], 2 ) == "OS") || (left( (string)v[i], 2 ) == "EW") ) { return i; }
		return findFactor(v, i);
	}

	i++;
	if ( i >= (signed int)v.size() ) { return i; }
	if (left( (string)v[i], 2 ) == "CW") { i++; return findFactor(v, i); }
	else { return i; }
}

vector<string> chasm_vocabulary::ccp_jrrTolkienizer(const string & s)
{
	// This is the J(on) R(elay) R(ocks) Tolkienizer!!!
	if (ccp_jrrTolkienList.empty())
	{
		//For this purpose, it's easier to keep track of what kind of token something is while you're tokenizing it.
		// OP = operator
		// OS = operator that's a symbol (doesn't need spaces around it)
		// SL = expression that includes 'short' or 'long' as an adjective
		// SK = throwaway word to skip over, or adjective that's unimportant
		// OB = object word (card, bkgnd, etc.)
		// CH = chunk word (char, word, line, etc.)
		// EW = word that ends an expression
		// CW = word that continues an expression
		// QU = quoted literal
		// ?? = unknown (most likely a variable, number, constant, misc keyword, etc.)
		ccp_jrrTolkienList.push_back("SLabbreviated version");
		ccp_jrrTolkienList.push_back("OPdoes not contain");
		ccp_jrrTolkienList.push_back("SLabbreviated date");
		ccp_jrrTolkienList.push_back("SLabbreviated time");
		ccp_jrrTolkienList.push_back("SLabbreviated name");
		ccp_jrrTolkienList.push_back("OPthere is not an");
		ccp_jrrTolkienList.push_back("OPdoesn't contain");
		ccp_jrrTolkienList.push_back("OPthere is not a");
		ccp_jrrTolkienList.push_back("OPthere isn't an");
		ccp_jrrTolkienList.push_back("SLabbrev version");
		ccp_jrrTolkienList.push_back("SLabbreviated id");
		ccp_jrrTolkienList.push_back("OPthere isn't a");
		ccp_jrrTolkienList.push_back("OPis not within");
		ccp_jrrTolkienList.push_back("SLshort version");
		ccp_jrrTolkienList.push_back("OPisn't within");
		ccp_jrrTolkienList.push_back("SLlong version");
		ccp_jrrTolkienList.push_back("SLabbr version");
		ccp_jrrTolkienList.push_back("OPthere is an");
		ccp_jrrTolkienList.push_back("SKabbreviated");
		ccp_jrrTolkienList.push_back("SLabbrev date");
		ccp_jrrTolkienList.push_back("SLabbrev time");
		ccp_jrrTolkienList.push_back("SLabbrev name");
		ccp_jrrTolkienList.push_back("OPthere is a");
		ccp_jrrTolkienList.push_back("OBbackground");
		ccp_jrrTolkienList.push_back("SLshort date");
		ccp_jrrTolkienList.push_back("SLshort time");
		ccp_jrrTolkienList.push_back("SLshort name");
		ccp_jrrTolkienList.push_back("OPis within");
		ccp_jrrTolkienList.push_back("OPis not in");
		ccp_jrrTolkienList.push_back("OPis not of");
		ccp_jrrTolkienList.push_back("OPis not an");
		ccp_jrrTolkienList.push_back("CHcharacter");
		ccp_jrrTolkienList.push_back("CHparagraph");
		ccp_jrrTolkienList.push_back("SLlong date");
		ccp_jrrTolkienList.push_back("SLlong time");
		ccp_jrrTolkienList.push_back("SLlong name");
		ccp_jrrTolkienList.push_back("SLabbr date");
		ccp_jrrTolkienList.push_back("SLabbr time");
		ccp_jrrTolkienList.push_back("SLabbr name");
		ccp_jrrTolkienList.push_back("SLabbrev id");
		ccp_jrrTolkienList.push_back("SKprevious");
		ccp_jrrTolkienList.push_back("OPcontains");
		ccp_jrrTolkienList.push_back("OPisn't in");
		ccp_jrrTolkienList.push_back("OPisn't of");
		ccp_jrrTolkienList.push_back("OPis not a");
		ccp_jrrTolkienList.push_back("OPisn't an");
		ccp_jrrTolkienList.push_back("CHsentence");
		ccp_jrrTolkienList.push_back("OBmenuItem");
		ccp_jrrTolkienList.push_back("OBvariable");
		ccp_jrrTolkienList.push_back("SLshort id");
		ccp_jrrTolkienList.push_back("OPisn't a");
		ccp_jrrTolkienList.push_back("CHelement");
		ccp_jrrTolkienList.push_back("EWwithout");
		ccp_jrrTolkienList.push_back("OBproject");
		ccp_jrrTolkienList.push_back("OBmessage");
		ccp_jrrTolkienList.push_back("OBwatcher");
		ccp_jrrTolkienList.push_back("SLlong id");
		ccp_jrrTolkienList.push_back("SLabbr id");
		ccp_jrrTolkienList.push_back("OPis not");
		ccp_jrrTolkienList.push_back("EWbefore");
		ccp_jrrTolkienList.push_back("OBbutton");
		ccp_jrrTolkienList.push_back("OBwindow");
		ccp_jrrTolkienList.push_back("SKabbrev");
		ccp_jrrTolkienList.push_back("OPis in");
		ccp_jrrTolkienList.push_back("OPis of");
		ccp_jrrTolkienList.push_back("OPis an");
		ccp_jrrTolkienList.push_back("OPisn't");
		ccp_jrrTolkienList.push_back("CHshort");
		ccp_jrrTolkienList.push_back("EWafter");
		ccp_jrrTolkienList.push_back("OBbkgnd");
		ccp_jrrTolkienList.push_back("OBstack");
		ccp_jrrTolkienList.push_back("OBfield");
		ccp_jrrTolkienList.push_back("OPis a");
		ccp_jrrTolkienList.push_back("CHchar");
		ccp_jrrTolkienList.push_back("CHword");
		ccp_jrrTolkienList.push_back("CHline");
		ccp_jrrTolkienList.push_back("CHitem");
		ccp_jrrTolkienList.push_back("CHpara");
		ccp_jrrTolkienList.push_back("CHbyte");
		ccp_jrrTolkienList.push_back("CHlong");
		ccp_jrrTolkienList.push_back("CHsent");
		ccp_jrrTolkienList.push_back("CHlist");
		ccp_jrrTolkienList.push_back("CHelem");
		ccp_jrrTolkienList.push_back("EWinto");
		ccp_jrrTolkienList.push_back("EWwith");
		ccp_jrrTolkienList.push_back("OBcard");
		ccp_jrrTolkienList.push_back("OBproj");
		ccp_jrrTolkienList.push_back("OBfile");
		ccp_jrrTolkienList.push_back("OBpart");
		ccp_jrrTolkienList.push_back("OBmenu");
		ccp_jrrTolkienList.push_back("SKthis");
		ccp_jrrTolkienList.push_back("SKprev");
		ccp_jrrTolkienList.push_back("SKnext");
		ccp_jrrTolkienList.push_back("SKabbr");
		ccp_jrrTolkienList.push_back("OS& &");
		ccp_jrrTolkienList.push_back("OS< =");
		ccp_jrrTolkienList.push_back("OS> =");
		ccp_jrrTolkienList.push_back("OS< >");
		ccp_jrrTolkienList.push_back("OPnot");
		ccp_jrrTolkienList.push_back("OPdiv");
		ccp_jrrTolkienList.push_back("OPmod");
		ccp_jrrTolkienList.push_back("OPand");
		ccp_jrrTolkienList.push_back("OBstk");
		ccp_jrrTolkienList.push_back("OBfld");
		ccp_jrrTolkienList.push_back("OBbtn");
		ccp_jrrTolkienList.push_back("OBURL");
		ccp_jrrTolkienList.push_back("OBmsg");
		ccp_jrrTolkienList.push_back("OBvar");
		ccp_jrrTolkienList.push_back("OBbox");
		ccp_jrrTolkienList.push_back("SKthe");
		ccp_jrrTolkienList.push_back("SKall");
		ccp_jrrTolkienList.push_back("OS&&");
		ccp_jrrTolkienList.push_back("OS<=");
		ccp_jrrTolkienList.push_back("OS>=");
		ccp_jrrTolkienList.push_back("OS<>");
		ccp_jrrTolkienList.push_back("OPis");
		ccp_jrrTolkienList.push_back("OPor");
		ccp_jrrTolkienList.push_back("EWto");
		ccp_jrrTolkienList.push_back("CWof");
		ccp_jrrTolkienList.push_back("CWin");
		ccp_jrrTolkienList.push_back("OBcd");
		ccp_jrrTolkienList.push_back("OBbg");
		ccp_jrrTolkienList.push_back("OS(");
		ccp_jrrTolkienList.push_back("OS)");
		ccp_jrrTolkienList.push_back("OS-");
		ccp_jrrTolkienList.push_back("OS^");
		ccp_jrrTolkienList.push_back("OS*");
		ccp_jrrTolkienList.push_back("OS/");
		ccp_jrrTolkienList.push_back("OS\\");
		ccp_jrrTolkienList.push_back("OS%");
		ccp_jrrTolkienList.push_back("OS|");
		ccp_jrrTolkienList.push_back("OS+");
		ccp_jrrTolkienList.push_back("OS&");
		ccp_jrrTolkienList.push_back("OS<");
		ccp_jrrTolkienList.push_back("OS>");
		ccp_jrrTolkienList.push_back("OS");
		ccp_jrrTolkienList.push_back("OS");
		ccp_jrrTolkienList.push_back("OS=");
		ccp_jrrTolkienList.push_back("OS");
		ccp_jrrTolkienList.push_back("OS,");
	}
	vector<string> v;
	vector<string>::iterator i;
	string ss = stripdupspaces(s);
	string t, ct, ctt;
	string operatorsymbols = " ()-^*/\\%|+&<>=,";
	bool fnd;
	bool nz;
	int o;
	while (ss != (string)"")
	{
		if (ss[0] == 34)
		{
			ss = mid(ss, 2);
			o = instr(ss, "\"");
			if ( o <= 0 )
			{
				v.push_back((string)"QU\"" + trim(ss));
				ss = (string)"";
			} else {
				v.push_back( (string)"QU\"" + trim(left(ss, o)) );
				ss = trim(mid(ss, o+1));
			}
		} else {
			fnd = false;
			for (i=ccp_jrrTolkienList.begin(); i<ccp_jrrTolkienList.end(); i++)
			{
				ct = (string)*i;
				ctt = mid(ct,3);
				if ( cie(left(ss, ctt.size()) , ctt) )
				{
					if ( (left(ct,2) == (string)"OS") || (instr(operatorsymbols, mid(ss,ctt.size()+1,1)) > 0) )
					{
						fnd = true;
						v.push_back(ct);
						ss = trim(mid(ss, ctt.size()+1));
						continue;
					}
				}
			}
			if (!fnd)
			{
				// o = instr(ss, " ");
				nz = false;
				o = 1;
				while ( (o<=(signed int)ss.size()) && (!nz) )
				{
					if (instr(operatorsymbols, mid(ss, o, 1)) > 0)
					{
						// DEBUGGING_CODE! :: cout << "operator symbol found: " << o << endl;
						nz = true;
						continue;
					} else { o++; }
				}
				if (!nz) { o = 0; }

				// DEBUGGING_CODE! :: cout << o << endl;

				if (o <= 0)
				{
					v.push_back((string)"??" + trim(ss));
					ss = (string)"";
				} else {
					v.push_back((string)"??" + trim(left(ss, o-1)));
					ss = trim(mid(ss, o));
				}
			}
		}
	}
	return v;
}

string chasm_vocabulary::ccp_jrrDeTolkienizer(const vector<string> & v)
{
	vector<string>::iterator i;
	string s;
	for (i=v.begin(); i<v.end(); i++)
	{
		s += (string)" ";
		s += mid( (string)*i ,3);
	}
	return mid(s, 2);
}

int chasm_vocabulary::mbwinstr(const string & inwhat, const string & findwhat)
{
	int i,o,m;
	string l;
	string r;
	if (instr(inwhat, findwhat) == 0) { return 0; }
	o = instr(inwhat, findwhat);
	m = inwhat.size() - findwhat.size() + 1;
	for (i=o; i<m; ++i)
	{
		if ( cie(mid(inwhat, i, findwhat.size()), findwhat) )
		{
			l = left(inwhat, i-1);
			r = mid(inwhat, i+1);
			if (equalparens(l) && equalquotes(l) && equalparens(r) && equalquotes(r))
			{
				return i;
			}
		}
	}
	return -1;
}

bool chasm_vocabulary::mbwsatisfied(const string & l, const string & pafter)
{
	int i, m, lastoffset;
	string paft;
	if ( (instr(pafter, "|")>0) && (pafter.size() != 1) ) {
		m = countfields(pafter, "|")+1;
		paft = "";
		for (i=1; i<m; ++i)
		{
			paft = " ";
			paft += trim(nthfield(pafter, "|", i));
			paft += " ";
			if (paft == "  ") {
				paft = " ";
			} else if ( (paft.size() == 3) && (trim(paft).size() == 1) ) {
				paft = trim(paft);
			}
			if (mbwinstr(l, paft)>0) { return true; }
		}
		return false;
	} else if ( (instr(pafter, "&")>0) && (pafter.size() != 1) ) {
		m = countfields(pafter, "&")+1;
		paft = "";
		for (i=1; i<m; ++i)
		{
			paft = " ";
			paft += trim(nthfield(pafter, "&", i));
			paft += " ";
			if (paft == "  ") {
				paft = " ";
			} else if ( (paft.size() == 3) && (trim(paft).size() == 1) ) {
				paft = trim(paft);
			}
			if (mbwinstr(l, paft)<1) { return false; }
		}
		return true;
	} else if ( (instr(pafter, "#")>0) && (pafter.size() != 1) ) {
		lastoffset = 0;
		m = countfields(pafter, "#")+1;
		paft = "";
		for (i=1; i<m; ++i)
		{
			paft = " ";
			paft += trim(nthfield(pafter, "#", i));
			paft += " ";
			if (paft == "  ") {
				paft = " ";
			} else if ( (paft.size() == 3) && (trim(paft).size() == 1) ) {
				paft = trim(paft);
			}
			if (mbwinstr(l, paft) < lastoffset) {
				return false;
			} else {
				lastoffset = mbwinstr(l, paft);
			}
		}
		return true;
	} else if ( (instr(pafter, "~")>0) && (pafter.size() != 1) ) {
		paft = " ";
		paft += trim(nthfield(pafter, "~", 1));
		paft += " ";
		if (! ((instr(l, paft)>0) | (paft == "  "))) { return false; }
		m = countfields(pafter, "~")+1;
		for (i=2; i<m; ++i)
		{
			paft = " ";
			paft += trim(nthfield(pafter, "~", i));
			paft += " ";
			if (paft == "  ") {
				paft = " ";
			} else if ( (paft.size() == 3) && (trim(paft).size() == 1) ) {
				paft = trim(paft);
			}
			if (mbwinstr(l, paft)>0) { return false; }
		}
		return true;
	} else {
		return ( (pafter == "  ") | (mbwinstr(l, pafter)>0) );
	}
}

void chasm_vocabulary::clear(void)
{
	wordz.clear();
}

string chasm_vocabulary::indextoname(int i)
{
	return wordz[i].name;
}

int chasm_vocabulary::namekindtoindex(const string & n, const string & k)
{
	vector<chasm_vocab_word>::iterator i;
	int j;
	for (i=wordz.begin(), j=0; i<wordz.end(); i++, j++)
	{
		if (cie(i->name, n) && cie(i->kind(), k)) { return j; }
	}
	return -1;
}

int chasm_vocabulary::nametoindex(const string & n)
{
	vector<chasm_vocab_word>::iterator i;
	int j;
	for (i=wordz.begin(), j=0; i<wordz.end(); i++, j++)
	{
		if (cie(i->name, n)) { return j; }
	}
	return -1;
}

int chasm_vocabulary::nametypetoindex(const string & n, const string & k)
{
	vector<chasm_vocab_word>::iterator i;
	int j;
	for (i=wordz.begin(), j=0; i<wordz.end(); i++, j++)
	{
		if (cie(i->name, n) && cie(i->type, k)) { return j; }
	}
	return -1;
}

void chasm_vocabulary::readdata(const string & s)
{
	chasm_vocab_word w;
	string d;
	string ss;
	string delim = " ";
	int o;
	delim[0] = 1;
	d = s;
	while ((d != "") && (d != delim))
	{
		o = instrb(d, delim);
		if (o <= 0) { o = d.size() + 1; }
		if (o == 1) {
			d = mid(d, 2);
		} else {
			ss = left(d, o-1);
			w.readdata(ss);
			wordz.push_back(w);
			d = mid(d, o+1);
		}
	}
}

void chasm_vocabulary::readfile(const string & filename)
{
	binarystream b(filename, false, false);
	if (b.ok()) {
		readdata(b.readStringValueAsString(b.getLength()));
		b.close();
	}
}

string chasm_vocabulary::writedata()
{
	string s = "";
	string delim = " ";
	vector<chasm_vocab_word>::iterator i;
	delim[0] = 1;
	for (i=wordz.begin(); i<wordz.end(); i++)
	{
		s += i->writedata();
		s += delim;
	}
	s = left(s, s.size()-1);
	return s;
}

void chasm_vocabulary::writefile(const string & filename)
{
	binarystream b(filename, true, true);
	string s = writedata();
	if (b.ok()) {
		b.write(s, s.size());
		b.close();
	}
}

const vector<chasm_vocab_word>::iterator chasm_vocabulary::begin(void)
{
	return wordz.begin();
}

const vector<chasm_vocab_word>::iterator chasm_vocabulary::end(void)
{
	return wordz.end();
}

void chasm_vocabulary::sort(void)
{
	bool sortd = false;
	int l = (int)wordz.size();
	int i;
	vector<chasm_vocab_word>::iterator ii;
	chasm_vocab_word v, vv;
	while (!sortd)
	{
		sortd = true;
		l--;
		for (i=0, ii=wordz.begin(); i<l; i++, ii++)
		{
			if ( (ii+1)->sortpriority() < ii->sortpriority() )
			{
				v = *ii;
				vv = *(ii+1);
				*ii = vv;
				*(ii+1) = v;
				sortd = false;
			} else if (  ( (ii+1)->sortpriority() == ii->sortpriority() )  &&  ( ((ii+1)->name).size() > (ii->name).size() )  )
			{
				v = *ii;
				vv = *(ii+1);
				*ii = vv;
				*(ii+1) = v;
				sortd = false;
			}
		}
	}
}

void chasm_vocabulary::sortbyname(void)
{
	bool sortd = false;
	int l = (int)wordz.size();
	int i;
	vector<chasm_vocab_word>::iterator ii;
	chasm_vocab_word v, vv;
	while (!sortd)
	{
		sortd = true;
		l--;
		for (i=0, ii=wordz.begin(); i<l; i++, ii++)
		{
			if ( ((ii+1)->name) < (ii->name) )
			{
				v = *ii;
				vv = *(ii+1);
				*ii = vv;
				*(ii+1) = v;
				sortd = false;
			}
		}
	}
}

vector<string> chasm_vocabulary::chasm_parse(const string & s)
{
	return chasm_parse(s, 0);
}

vector<string> chasm_vocabulary::chasm_parse(const string & s, int parsingMode)
{
	/* esta function usa muchas variables */
	string scr, oscr, c;
	vector<chasm_vocab_word>::iterator i;
	vector<string> cc;
	vector<string> unnumberedparams;
	bchasm bch;
	stack<variant> bchparams;
	vector<string>::iterator pi;
	string pafter, pkeywd, pbefore, pname;
	string pkeywdp;
	int pnum, pnum2, pselfnum;
	bool pincrement, pcondense, pbalancechek;
	string pdefault;
	string pkeywdreset;
	unsigned long int o, oi, oo;
	string l, r, ll, rr;
	int pmi;
	vector<string> pmn(255, "");
	vector<int> pmnum(255, 0);
	vector<string> pmdef(255, "");
	string bpn;
	int bpnum;
	string bpdef;
	string plist, htequ, pvar, pval;
	vector<string> stparams(36, "");
	vector<bool> stparamexists(36, false);
	hasher h;
	int ppi, ppm;
	string thecomment = "";
	string s1, s2;
	string alphabet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz_";
	string alphanumbet = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789_";
	string optionl = " ";
	string quot = " ";
	string pdelim = "  ";
	string pdelims = " ";
	string cr = " ";
	optionl[0] = 194;
	quot[0] = 34;
	pdelim[0] = 127;
	pdelim[1] = 127;
	pdelims[0] = 127;
	cr[0] = 13;
	if ( (parsingMode < 0) || (parsingMode > 2) )
	{
		cc.push_back( (string)"#PARSEERROR: Invalid parsing mode." );
		return cc;
	}
	if (stripSpaces)
	{
		scr = stripdupspaces(s);
	} else {
		scr = trim(s);
	}
	if (instr(scr, "--")>0)
	{
		o = instr(scr, "--");
		for (oi = o; oi <= scr.size(); oi++)
		{
			if (mid(scr, oi, 2) == "--")
			{
				l = left(scr, oi-1);
				r = mid(scr, oi+2);
				if ( equalquotes(l) && equalparens(l) )
				{
					thecomment = trim(r);
					scr = l;
					continue;
				}
			}
		}
	}
	if (convertoptionl && instr(scr, optionl)>0)
	{
		o = instr(scr, optionl);
		for (oi = scr.size(); oi >= 0; oi--)
		{
			if (mid(scr, oi, 1) == optionl)
			{
				l = left(scr, oi-1);
				r = mid(scr, oi+2);
				if (! equalquotes(l))
				{
					/* s1 = mid(scr, oi+1, 1); */
					switch(toupper(scr[oi])){
					case 'Q':
						scr = l + quot + (string)"&quote&" + quot + r;
						break;
					case 'R':
						scr = l + quot + (string)"&return&" + quot + r;
						break;
					case 'L':
						scr = l + quot + (string)"&linefeed&" + quot + r;
						break;
					case 'F':
						scr = l + quot + (string)"&formfeed&" + quot + r;
						break;
					case 'T':
						scr = l + quot + (string)"&tab&" + quot + r;
						break;
					case 'E':
						scr = l + quot + (string)"&(the numtochar of 27)&" + quot + r;
						break;
					case 'D':
						scr = l + quot + (string)"&(the numtochar of 127)&" + quot + r;
						break;
					case 'B':
						scr = l + quot + (string)"&(the numtochar of 8)&" + quot + r;
						break;
					case 'V':
						scr = l + quot + (string)"&(the numtochar of 11)&" + quot + r;
						break;
					case 'P':
						scr = l + quot + (string)"&pi&" + quot + r;
						break;
					case (194 /* option-L; change this if you change optionl above */):
						scr = l + optionl + r;
						break;
					default:
						scr = l + r;
						break;
					}
				}
			}
		}
	}
	scr = trim(scr) + " ";
	if ( processAssignment && (instr(scr, (string)"=") > 0) )
	{
		o = instr(scr, (string)"=");
		oi = 1;
		l = trim(left(scr, o-1));
		r = trim(mid(scr, o+1));
		if ( instr(alphabet, left(l, 1)) > 0 )
		{
			for (o = 1; o <= l.size(); o++)
			{
				if ( instr(alphanumbet, mid(l, o, 1)) <= 0 ) { oi = 0; }
			}
			if ( oi == 1 )
			{
				scr = (string)"let " + l + (string)" be " + r;
			}
		}
	}
	if ( processFunctiOn )
	{
		if ( ( left(scr, 3) == (string)"on " ) || ( left(scr, 9) == (string)"function " ) )
		{
			cc.clear();
			s1 = nthfield(scr, " ", 1);
			s2 = nthfield(scr, " ", 2);
			scr = trim(mid(scr, s1.size()+s2.size()+3));
			if ( (parsingMode == 0) || (parsingMode == 2) )
			{
				cc.push_back(s1);
				cc.push_back(s2);
				if (scr != (string)"")
				{
					oo = countfields(scr, (string)",");
					for (o = 1; o <= oo; o++)
					{
						cc.push_back( trim(nthfield(scr, (string)",", o)) );
					}
				}
			}
			if ( (parsingMode == 1) || (parsingMode == 2) )
			{
				cc.push_back((string)":command");
				cc.push_back(s1);
				cc.push_back((string)":commandFirst");
				cc.push_back(s1);
				cc.push_back((string)":commandRest");
				cc.push_back((string)"");
				cc.push_back((string)":comment");
				cc.push_back(thecomment);
				cc.push_back((string)":handlerName");
				cc.push_back(s2);
				if (scr != (string)"")
				{
					oo = countfields(scr, (string)",");
					for (o = 1; o <= oo; o++)
					{
						if (o == 1) { cc.push_back((string)":firstParam"); }
						else if (o == 2) { cc.push_back((string)":nextParam"); }
						else if (o > 2) { cc.push_back((string)":nextParam" + str(o-1)); }
						cc.push_back( trim(nthfield(scr, (string)",", o)) );
					}
				}
			}
			return cc;
		}
	}
	oscr = scr;
	for (i = wordz.begin(); i < wordz.end(); i++)
	{
		c = i->name + " ";
		if ( c == (string)" " ) { c = (string)""; }
		if (cie(left(scr,c.size()), c))
		{
			s2 = i->type;
			h.char_val[0] = s2[0];
			h.char_val[1] = s2[1];
			h.char_val[2] = s2[2];
			h.char_val[3] = s2[3];
			switch(h.hash_val){
			case 'chas': /* CHASM assembly */
				while (!bchparams.empty()) { bchparams.pop(); }
				bchparams.push(variant(scr));
				bch.assemble( i->gparams(0) );
				scr = bch.interpret_with_prams( bchparams );
				break;
			case 'rbsc': /* RBScript */
				break;
			case 'ccom': /* CHASM Command */
				cc.clear();
				s1 = nthfield(trim(i->name)," ",1);
				if ( (parsingMode == 0) || (parsingMode == 2) )
				{
					cc.push_back(s1);
				}
				cc.resize(255, (string)"");
				unnumberedparams.clear();
				if ( (parsingMode == 1) || (parsingMode == 2) )
				{
					// DEBUGGING_CODE! :: cout << "Defining Command" << endl;
					unnumberedparams.push_back((string)":command");
					unnumberedparams.push_back(i->name);
					// DEBUGGING_CODE! :: cout << "Defining CommandFirst" << endl;
					unnumberedparams.push_back((string)":commandFirst");
					unnumberedparams.push_back(s1);
					// DEBUGGING_CODE! :: cout << "Defining CommandRest" << endl;
					unnumberedparams.push_back((string)":commandRest");
					s1 = trim(i->name);
					s2 = nthfield(s1, " ", 1);
					o = s2.size();
					s1 = trim(mid(s1, o+1));
					unnumberedparams.push_back(s1);
					// DEBUGGING_CODE! :: cout << "Defining Comment" << endl;
					unnumberedparams.push_back((string)":comment");
					unnumberedparams.push_back(thecomment);
					// DEBUGGING_CODE! :: cout << "Reached End of Defining Command* Block" << endl;
				}
				s1 = trim(i->name);
				s2 = nthfield(s1, " ", 1);
				o = s2.size();
				s1 = trim(mid(s1, o+1));
				pmi = 1;
				// DEBUGGING_CODE! :: cout << "CHASM Command Found: " << i->name << endl;
				for (pi = i->params.begin(); pi < i->params.end(); pi+=4)
				{
					pafter = (string)" " + trim(*pi) + (string)" ";
					pbefore = (string)" " + trim(*(pi+1)) + (string)" ";
					pkeywd = *(pi+2);
					pname = *(pi+3);
					// DEBUGGING_CODE! :: cout << "Processing Keyword: " << pkeywd << " (#" << pnum << ")" << endl;

					/* NEW FORMAT OF PNAME INTRODUCED IN CHASM C/P 2.0:
					a#b#c#d#e, where:
					a = name given to the parameter
					b = number given to the parameter
					c = '+' if the number should be incremented for multiple parameters
					d = number of the parameter holding the keyword itself, if applicable
					e = value of the parameter if it is not passed
					If a == '@', the value of the parameter is the command itself,
					and c and d are the range of words in the command to put into the parameter,
					with c = -1 if we want just the first word or c = 0 if we want all but the first word. */

					pnum = (int)val(nthfield(pname, (string)"#", 2));
					pnum2 = (int)val(nthfield(pname, (string)"#", 3));
					pincrement = (instr(nthfield(pname, (string)"#", 3), (string)"+")>0);
					pbalancechek = (instr(nthfield(pname, (string)"#", 3), (string)"B")>0);
					pcondense = (pcondense | (instr(nthfield(pname, (string)"#", 3), (string)"C")>0));
					pselfnum = (int)val(nthfield(nthfield(pname, (string)"#", 4), (string)"*", 1));
					pkeywdreset = nthfield(nthfield(pname, (string)"#", 4), (string)"*", 2);
					pdefault = nthfield(pname, (string)"#", 5);
					pname = nthfield(pname, (string)"#", 1);
					// DEBUGGING_CODE! :: cout << "PKeywdReset: " << pkeywdreset.c_str() << endl;

					if (pkeywd == (string)"")
					{
						// DEBUGGING_CODE! :: cout << "(Blank Parameter)" << endl;
						bpn = pname;
						bpnum = pnum;
						bpdef = pdefault;
					}
					else if (pkeywd == (string)"@")
					{
						// DEBUGGING_CODE! :: cout << "(@ Parameter)" << endl;
						s1 = i->name;
						if (pnum2 == -1)
						{
							s1 = nthfield(s1, " ", 1);
						}
						else if (pnum2 == 0)
						{
							s1 = trim(s1);
							o = nthfield(s1, " ", 1).size();
							s1 = trim(mid(s1, o+1));
						}
						else if (pnum2 >= 1)
						{
							if (pselfnum <= pnum2)
							{
								s1 = nthfield(s1, " ", pnum2);
							} else {
								s1 = "";
								for (o = (unsigned int)pnum2; o <= (unsigned int)pselfnum; o++)
								{
									s1 = s1 + nthfield(i->name, " ", o) + " ";
								}
								s1 = trim(s1);
							}
						}
						if ((trim(s1) == (string)"") && (pdefault != (string)"")) { s1 = pdefault; }
						// DEBUGGING_CODE! :: cout << "Value: " << s1.c_str() << endl;
						if ( ((parsingMode == 0) || (parsingMode == 2)) && (pnum != 0))
						{
							// DEBUGGING_CODE! :: cout << "Number Given: " << pnum << endl;
							cc[pnum] = s1;
							if (parsingMode == 2)
							{
								unnumberedparams.push_back((string)":" + pname);
								unnumberedparams.push_back(s1);
							}
						}
						else if ((pname != (string)"n/a") && (pname != (string)""))
						{
							unnumberedparams.push_back((string)":" + pname);
							unnumberedparams.push_back(s1);
						}
					}
					else if ((pkeywd.size() == 1) && (instr(pkeywd, (string)"abcdefghijklmnopqrstuvwxyz") <= 0))
					{
						// DEBUGGING_CODE! :: cout << "(One-Char Kywd Parameter)" << endl;
						o = instr(scr, pkeywd);
						if (o>0)
						{
							for (oi = o; oi <= scr.size(); oi++)
							{
								if (scr[oi-1] == pkeywd[0])
								{
									oo = instr(oscr, pkeywd) + (oi-o);
									l = left(scr, oi-1);
									r = mid(scr, oi+1);
									ll = left(oscr, oo-1);
									rr = mid(oscr, oi+1);
									// DEBUGGING_CODE! :: cout << "Found in Text" << endl << "   Pos: " << oo << endl << "   L: " << l.c_str() << endl << "   R: " << r.c_str() << endl;
									// DEBUGGING_CODE! :: cout << "   LL: " << ll.c_str() << endl << "   RR: " << rr.c_str() << endl;
									if (mbwsatisfied(ll, pafter) && mbwsatisfied(rr, pbefore))
									{
										if ( equalparens(l) && equalquotes(l) && ((!pbalancechek) || balancedexpr(l)) )
										{
											// DEBUGGING_CODE! :: cout << "   Conditions Satisfied (PAft, PBef, = Parens, Quotes)" << endl;
											scr = l + (string)" " + pdelim + chr(pmi) + (string)" " + r;
											pmn[pmi] = pname;
											pmnum[pmi] = pnum;
											pmdef[pmi] = pdefault;
											if (pincrement) { pnum++; }
											if ( ((parsingMode == 0) || (parsingMode == 2)) && (pselfnum > 0) )
											{
												// DEBUGGING_CODE! :: cout << "   Num Given to Keyword: " << pselfnum << endl;
												cc[pselfnum] = pname;
											}
											if (instr((string)"0123456789", right(pname,1)) <= 0)
											{
												pname += (string)"2";
											} else {
												s1 = "";
												while ( instr((string)"0123456789", right(pname,1)) > 0 )
												{
													s1 = right(pname, 1) + s1;
													pname = left(pname, pname.size()-1);
												}
												pname += str(val(s1)+1);
											}
											pmi++;
										}
									}
								}
							}
						}
					}
					else
					{
						// DEBUGGING_CODE! :: cout << "(Multi-Char Kywd Parameter)" << endl;
						pkeywdp = (string)" " + trim(pkeywd) + (string)" ";
						o = instr(scr, pkeywdp);
						// DEBUGGING_CODE! :: cout << "o=" << o << endl;
						// DEBUGGING_CODE! :: cout << "pkeywdp=" << pkeywdp << endl;
						if ((signed int)o>0)
						{
							for (oi = o; oi <= (scr.size() - pkeywd.size()); oi++)
							{
								// DEBUGGING_CODE! :: cout << "current stuff=" << mid(scr, oi, pkeywdp.size()) << endl;
								if ( cie( mid(scr, oi, pkeywdp.size()), pkeywdp ) )
								{
									oo = instr(oscr, pkeywdp) + (oi-o);
									l = left(scr, oi-1);
									r = mid(scr, oi+pkeywdp.size());
									ll = left(oscr, oo-1);
									rr = mid(oscr, oo+pkeywdp.size());
									// DEBUGGING_CODE! :: cout << "Found in Text" << endl << "   Pos: " << oo << endl << "   L: " << l.c_str() << endl << "   R: " << r.c_str() << endl;
									// DEBUGGING_CODE! :: cout << "   LL: " << ll.c_str() << endl << "   RR: " << rr.c_str() << endl;
									if (mbwsatisfied(ll, pafter) && mbwsatisfied(rr, pbefore))
									{
										// DEBUGGING_CODE! :: cout << "   PAfter/PBefore Satisfied" << endl;
										if ( equalparens(l) && equalquotes(l) && ((!pbalancechek) || balancedexpr(l)) )
										{
											// DEBUGGING_CODE! :: cout << "   Conditions Satisfied (PAft, PBef, = Parens, Quotes)" << endl;
											scr = l + (string)" " + pdelim + chr(pmi) + (string)" " + r;
											pmn[pmi] = pname;
											pmnum[pmi] = pnum;
											pmdef[pmi] = pdefault;
											if (pincrement) { pnum++; }
											if ( ((parsingMode == 0) || (parsingMode == 2)) && (pselfnum > 0) )
											{
												// DEBUGGING_CODE! :: cout << "   Num Given to Keyword: " << pselfnum << endl;
												if (pkeywdreset == (string)"")
												{
													cc[pselfnum] = pkeywd;
												} else {
													cc[pselfnum] = pkeywdreset;
												}
											}
											if (instr((string)"0123456789", right(pname,1)) <= 0)
											{
												pname += (string)"2";
											} else {
												s1 = "";
												while ( instr((string)"0123456789", right(pname,1)) > 0 )
												{
													s1 = right(pname, 1) + s1;
													pname = left(pname, pname.size()-1);
												}
												pname += str(val(s1)+1);
											}
											pmi++;
										}
									}
								}
							}
						}
					}
				}
				s1 = trim(mid(nthfield(scr, pdelim, 1), c.size()));
				if ( (s1 == (string)"") && (bpdef != (string)"") ) { s1 = bpdef; }
				// DEBUGGING_CODE! :: cout << "S1: " << s1 << endl;
				if ( ((parsingMode == 0) || (parsingMode == 2)) && (bpnum != 0) )
				{
					// DEBUGGING_CODE! :: cout << "Num Given to S1: " << bpnum << endl;
					cc[bpnum] = s1;
					if (parsingMode == 2)
					{
						unnumberedparams.push_back((string)":" + bpn);
						unnumberedparams.push_back(s1);
					}
				}
				else if ( (bpn != (string)"n/a") && (bpn != (string)"") )
				{
					unnumberedparams.push_back((string)":" + bpn);
					unnumberedparams.push_back(s1);
				}
				ppm = countfields(scr, pdelim);
				// DEBUGGING_CODE! :: cout << "Number of Fields: " << ppm << endl;
				if (ppm > 1)
				{
					// DEBUGGING_CODE! :: cout << "Delimited String OK" << endl;
					for (ppi = 2; ppi <= ppm; ppi++)
					{
						l = nthfield(scr, pdelim, ppi);
						s1 = trim(mid(l,2));
						o = left(l,1).c_str()[0];
						// DEBUGGING_CODE! :: cout << "Field " << ppi << endl << "   Content: " << s1 << endl << "   Param Num: " << o << endl;
						if ( (s1 == (string)"") && (pmdef[o] != (string)"") )
						{
							s1 = pmdef[o];
							// DEBUGGING_CODE! :: cout << "   Default Used: " << s1 << endl;
						}
						pmi = pmnum[o];
						s2 = pmn[o];
						if ( ((parsingMode == 0) || (parsingMode == 2)) && (pmi > 0) )
						{
							// DEBUGGING_CODE! :: cout << "   <Giving Num to Param - " << pmi << ">" << endl;
							cc[pmi] = s1;
							// DEBUGGING_CODE! :: cout << "   Num Given to Parameter: " << pmi << endl;
							if (parsingMode == 2)
							{
								unnumberedparams.push_back((string)":" + s2);
								unnumberedparams.push_back(s1);
							}
						}
						else if ( (s2 != (string)"n/a") && (s2 != (string)"") )
						{
							unnumberedparams.push_back((string)":" + s2);
							unnumberedparams.push_back(s1);
						}
					}
				}
				if (! cc.empty())
				{
					while ( (! cc.empty()) && (cc.back() == (string)"") )
					{
						cc.pop_back();
					}
					if ( pcondense && (! cc.empty()) )
					{
						for (pi = cc.end()-1; pi >= cc.begin(); pi--)
						{
							if (*pi == (string)"")
							{
								cc.erase(pi);
							}
						}
					}
				}
				while (! unnumberedparams.empty())
				{
					pi = unnumberedparams.begin();
					s2 = *pi;
					cc.push_back(s2);
					unnumberedparams.erase(pi);
				}
				return cc;
				break;
			case 'comm': /* SphereTalk Command */
				// DEBUGGING_CODE! :: cout << "Found STC" << endl;
				plist = i->gparams(0) + cr + i->gparams(1);
				htequ = i->htequiv();
				scr = mid(scr,c.size()+1);
				ppm = countfields(plist, cr);
				// DEBUGGING_CODE! :: cout << " Starting Keywd Search Loop" << endl;
				for (ppi = 1; ppi<=ppm; ppi++)
				{
					pkeywd = nthfield(plist, cr, ppi);
					pkeywd = trim( left( pkeywd, pkeywd.size()-nthfield(pkeywd," ",countfields(pkeywd," ")).size() ) );
					pkeywd = (string)" " + pkeywd + (string)" ";
					if (instr(scr,pkeywd) > 0)
					{
						o = instr(scr,pkeywd)+1;
						oo = o+pkeywd.size()-2;
						pvar = nthfield(plist, cr, ppi);
						pvar = mid(nthfield(pvar, " ", countfields(pvar, " ")), 2, 1);
						scr = left(scr, o-1) + pdelim + pvar + mid(scr, oo+1);
					}
				}
				stparams[0] = trim(nthfield(scr, pdelim, 1));
				stparamexists[0] = true;
				ppm = countfields(scr, pdelim);
				// DEBUGGING_CODE! :: cout << " Starting Param Ret Loop" << endl;
				if (ppm > 1)
				{
					for (ppi = 2; ppi<=ppm; ppi++)
					{
						pvar = left(nthfield(scr, pdelim, ppi),1);
						pval = trim(mid(nthfield(scr, pdelim, ppi), 2));
						s2 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
						o = instr(s2,pvar)-1;
						if (o>0)
						{
							stparams[o] = pval;
							stparamexists[o] = true;
						}
					}
				}
				s1 = stparams[0];
				// DEBUGGING_CODE! :: cout << " Starting Param(0) Split Loop" << endl;
				for (ppi = 0; ppi<(signed int)s1.size(); ppi++)
				{
					if (s1[ppi] == ',')
					{
						l = left(s1,ppi);
						r = mid(s1,ppi+2);
						if (equalparens(l) && equalquotes(l))
						{
							s1 = l + pdelim + r;
						}
					}
				}
				// DEBUGGING_CODE! :: cout << " Starting Param(0) Ret Loop" << endl;
				for (ppi = 9; ppi>=0; ppi--)
				{
					stparams[ppi] = nthfield(s1, pdelim, ppi+1);
					stparamexists[ppi] = (stparams[ppi] != (string)"");
				}
				scr = htequ;
				// DEBUGGING_CODE! :: cout << " Starting HTEqu Decode Loop" << endl;
				for (ppi = 0; ppi<(signed int)scr.size(); ppi++)
				{
					s2 = " ";
					s2[0] = scr[ppi];
					if ( (s2 == (string)"@") || (s2 == (string)"#") || (s2 == (string)"%") || ((unsigned char)s2[0] == 228) )
					{
						// DEBUGGING_CODE! :: cout << "  Found @#% at " << ppi << endl;
						l = left(scr, ppi);
						r = mid(scr, ppi+3);
						// DEBUGGING_CODE! :: cout << "  L: " << l << endl;
						// DEBUGGING_CODE! :: cout << "  R: " << r << endl;
						if (equalparens(l) && equalquotes(l))
						{
							// DEBUGGING_CODE! :: cout << "  Verified @#%" << endl;
							s1 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
							s2 = " ";
							s2[0] = scr[ppi+1];
							o = instr(s1, s2)-1;
							switch ((unsigned char)scr[ppi]) {
							case '%':
								scr = l + stparams[o] + r;
								break;
							case '#':
								if (stparamexists[o])
								{
									scr = l + (string)"true" + r;
								} else {
									scr = l + (string)"false" + r;
								}
								break;
							case 228:
								if (stparamexists[o])
								{
									scr = l + (string)"true" + r;
								} else {
									scr = l + (string)"false" + r;
								}
								break;
							case '@':
								scr = l + quot + replaceall( replaceall( stparams[o], quot, pdelim + (string)"&quote&" + pdelim ), pdelim, quot ) + quot + r;
								break;
							}
						}
					}
				}
				break;
			}
		}

		s2 = i->type;
		h.char_val[0] = s2[0];
		h.char_val[1] = s2[1];
		h.char_val[2] = s2[2];
		h.char_val[3] = s2[3];
		switch(h.hash_val){
		case 'func': /* SphereTalk Function */
			htequ = i->htequiv();
			pname = i->name;
			s1 = (string)"the "+pname;
			ll = " (&=<>+-*/^\\%|,";
			rr = " )&=<>+-*/^\\%|,";
			o = instr(scr, s1);
			if (o > 0)
			{
				if (instr(htequ, (string)"%0") > 0) /* function takes an argument */
				{
					s1 += (string)" of ";
					o = instr(scr, s1);
					if (o > 0)
					{
						for (oi = o; oi<scr.size(); oi++)
						{
							if ( cie( mid(scr, oi, s1.size()), s1 ) )
							{
								if ( instr(ll, mid(scr, oi-1, 1)) > 0 )
								{
									l = left(scr, oi-1);
									r = mid(scr, oi+s1.size());
									/* now the tricky part is finding the end of the argument... */
									//for (oo = 1; oo<=r.size(); oo++) /*hyp*/
									//{
									//	if ( instr(rr, mid(r,oo,1)) > 0 ) /*ert*/
									//	{
									//		s2 = trim(left(r, oo-1)); /*al*/
									//		oscr = trim(mid(r, oo+1)); /*k is*/
									//		if ( equalparens(s2) && equalquotes(s2) && balancedexpr(s2) ) /*a bi*/
									//		{
									//			pafter = nthfield(s2, " ", countfields(s2, " ")); /*tch*/
									//			pbefore = nthfield(oscr, " ", 1); /*to t*/
									//			if ( (pafter != (string)"the") && (pafter != (string)"of") && (pbefore != (string)"of") &&
									//			(pbefore != (string)"cd") && (pbefore != (string)"card") && (pbefore != (string)"bg") &&
									//			(pbefore != (string)"bkgnd") && (pbefore != (string)"background") && (pbefore != (string)"stk") &&
									//			(pbefore != (string)"stack") && (pbefore != (string)"proj") && (pbefore != (string)"project") &&
									//			(pbefore != (string)"wind") && (pbefore != (string)"window") && (pbefore != (string)"file") &&
									//			(pbefore != (string)"btn") && (pbefore != (string)"button") && (pbefore != (string)"fld") &&
									//			(pbefore != (string)"field") && (pbefore != (string)"URL") && (pbefore != (string)"part") &&
									//			(pafter != (string)"cd") && (pafter != (string)"card") && (pafter != (string)"bg") &&
									//			(pafter != (string)"bkgnd") && (pafter != (string)"background") && (pafter != (string)"stk") &&
									//			(pafter != (string)"stack") && (pafter != (string)"proj") && (pafter != (string)"project") &&
									//			(pafter != (string)"wind") && (pafter != (string)"window") && (pafter != (string)"file") &&
									//			(pafter != (string)"btn") && (pafter != (string)"button") && (pafter != (string)"fld") &&
									//			(pafter != (string)"field") && (pafter != (string)"URL") && (pafter != (string)"part") &&
									//			(pafter != (string)"short") && (pafter != (string)"long") && (pafter != (string)"abbr") &&
									//			(pafter != (string)"abbrev") && (pafter != (string)"abbreviated") )
									//			{
									//				s2 = left(r, oo-1); /*ok*/
									//				r = mid(r, oo); /*en*/
									//				continue; /*ize*/
									//			}
									//		}
									//	}
									//}
									s2 = findFactor(r);
									r = nthfield(s2, pdelims, 2);
									s2 = nthfield(s2, pdelims, 1);
									/* when we're done with it we should put the argument into s2, and all the stuff after into r */
									s2 = replaceall(htequ, (string)"%0", s2);
									s2 = replaceall(s2, (string)"#0", (string)"true");
									s2 = replaceall(s2, (string)"0", (string)"true");
									scr = l + s2 + r;
								}
							}
						}
					}
				}
				else /* function doesn't take an argument */
				{
					for (oi = o; oi<scr.size(); oi++)
					{
						if ( cie( mid(scr, oi, s1.size()), s1 ) )
						{
							if ( instr(ll, mid(scr, oi-1, 1)) > 0 )
							{
								if ( instr(rr, mid(scr, oi+s1.size(), 1)) > 0 )
								{
									l = left(scr, oi-1);
									r = mid(scr, oi+s1.size());
									scr = l + htequ + r;
								}
							}
						}
					}
				}
			}
			break;
		case 'sett': /* SphereTalk Property Setter */
			if ( left(scr,4) == (string)"set " ) {
				htequ = i->htequiv();
				pname = i->name;
				pkeywd = i->gparams(0);
				o = 0;
				s1 = scr;
				s2 = (string)" to " + pkeywd + (string)" ";
				for (ppi = 1; ppi <= (signed int)s1.size(); ppi++)
				{
					if ( cie(mid(s1, ppi, s2.size()), s2) )
					{
						l = left(s1, ppi-1);
						r = mid(s1, ppi+s2.size());
						if ( equalparens(l) && equalquotes(l) && balancedexpr(l) && ((o%2) == 0) )
						{
							s1 = l + pdelim + r;
							o++;
						}
					}
					else if ( cie(mid(s1, ppi, 4), (string)" to ") )
					{
						l = left(s1, ppi-1);
						r = mid(s1, ppi+4);
						if ( equalparens(l) && equalquotes(l) && balancedexpr(l) && ((o%2) == 0) )
						{
							s1 = l + pdelim + r;
							o++;
						}
					}
					else if ( cie(mid(s1, ppi, 4), (string)" of ") )
					{
						l = left(s1, ppi-1);
						r = mid(s1, ppi+4);
						if ( equalparens(l) && equalquotes(l) && balancedexpr(l) && (o == 0) )
						{
							s1 = l + pdelim + r;
							o+=2;
						}
					}
				}
				s2 = trim(nthfield(s1,pdelim,1));
				// DEBUGGING_CODE! :: cout << "s2: " << s2 << endl;
				if ( cie(s2,((string)"set the "+pname)) || cie(s2,((string)"set "+pname)) )
				{
					s2 = trim(nthfield(s1,pdelim,2)); /* this is either the value or the object */
					s1 = trim(nthfield(s1,pdelim,3)); /* this is the value, if an object was specified */
					if ( s1 == (string)"" )
					{
						stparams[0] = "";
						stparams[1] = s2;
						stparamexists[0] = false;
						stparamexists[1] = true;
					} else {
						stparams[0] = s2;
						stparams[1] = s1;
						stparamexists[0] = true;
						stparamexists[1] = true;
					}
					scr = htequ;
					// DEBUGGING_CODE! :: cout << " Starting HTEqu Decode Loop" << endl;
					for (ppi = 0; ppi<(signed int)scr.size(); ppi++)
					{
						s2 = " ";
						s2[0] = scr[ppi];
						if ( (s2 == (string)"@") || (s2 == (string)"#") || (s2 == (string)"%") || ((unsigned char)s2[0] == 228) )
						{
							// DEBUGGING_CODE! :: cout << "  Found @#% at " << ppi << endl;
							l = left(scr, ppi);
							r = mid(scr, ppi+3);
							// DEBUGGING_CODE! :: cout << "  L: " << l << endl;
							// DEBUGGING_CODE! :: cout << "  R: " << r << endl;
							if (equalparens(l) && equalquotes(l))
							{
								// DEBUGGING_CODE! :: cout << "  Verified @#%" << endl;
								s1 = "0123456789ABCDEFGHIJKLMNOPQRSTUVWXYZ";
								s2 = " ";
								s2[0] = scr[ppi+1];
								o = instr(s1, s2)-1;
								switch ((unsigned char)scr[ppi]) {
								case '%':
									scr = l + stparams[o] + r;
									break;
								case '#':
									if (stparamexists[o])
									{
										scr = l + (string)"true" + r;
									} else {
										scr = l + (string)"false" + r;
									}
									break;
								case 228:
									if (stparamexists[o])
									{
										scr = l + (string)"true" + r;
									} else {
										scr = l + (string)"false" + r;
									}
									break;
								case '@':
									scr = l + quot + replaceall( replaceall( stparams[o], quot, pdelim + (string)"&quote&" + pdelim ), pdelim, quot ) + quot + r;
									break;
								}
							}
						}
					}
				}
			}
			break;
		case 'cons': /* SphereTalk Constant */
			pval = i->htequiv();
			pvar = i->name;
			s1 = " (&=<>+-*/^\\%|,";
			s2 = " )&=<>+-*/^\\%|,";
			// DEBUGGING_CODE! :: cout << "Searching for Constant: " << pvar << endl;
			if ( instr(scr, pvar) > 0 )
			{
				// DEBUGGING_CODE! :: cout << "Found Constant: " << pvar << endl;
				scr = (string)" " + trim(scr) + (string)" ";
				for (ppi = 1; ppi <= (signed int)scr.size(); ppi++)
				{
					l = mid(scr, ppi, pvar.size());
					// DEBUGGING_CODE! :: cout << ": " << l << endl;
					if (cie(l,pvar))
					{
						// DEBUGGING_CODE! :: cout << "Index: " << ppi << endl;
						// DEBUGGING_CODE! :: cout << "L Char: " << mid(scr, ppi-1, 1) << endl;
						// DEBUGGING_CODE! :: cout << "R Char: " << mid(scr, ppi+pvar.size(), 1) << endl;
						if ( (ppi == 1) || (instr(s1, mid(scr, ppi-1, 1)) > 0) )
						{
							if ( (ppi == (signed int)scr.size()) || (instr(s2, mid(scr, ppi+pvar.size(), 1)) > 0) )
							{
								l = left(scr, ppi-1);
								r = mid(scr, ppi + pvar.size());
								// DEBUGGING_CODE! :: cout << "L: " << l << endl << "R: " << r << endl;
								if ( equalquotes(l) && equalquotes(r) )
								{
									scr = l + pval + r;
									// DEBUGGING_CODE! :: cout << "Scr: " << scr << endl;
								}
							}
						}
					}
				}
				scr = trim(scr);
			}
			break;
		}
	}
	/* no matching vocabulary was found, so it's a custom handler call */
	cc.clear();
	cc.push_back(nthfield(scr, " ", 1));
	scr=trim(mid(scr,nthfield(scr," ",1).size()+2));
	oo = 1;
	for (oi = 1; oi <= scr.size(); oi++)
	{
		if (scr[oi-1] == ',')
		{
			l = left(scr, oi-1);
			if (equalparens(l) && equalquotes(l))
			{
				cc.push_back(trim(mid(l,oo)));
				oo = oi+1;
			}
		}
	}
	cc.push_back(trim(mid(scr,oo)));
	return cc;
}


void ccp_init(void)
{
	ccp_global_vocab.clear();
	ccp_global_vocab.readfile("Standard Vocabulary");
	ccp_global_vocab.sort();
}

vector<string> ccp_parse(const string & s)
{
	return ccp_global_vocab.chasm_parse(s, 0);
}

vector<string> ccp_parse(const string & s, int m)
{
	return ccp_global_vocab.chasm_parse(s, m);
}

map<string,string> ccp_vector_to_map(const vector<string> & v)
{
	map<string,string> m;
	vector<string>::iterator i;
	string cs, pp; // current string, previous parameter
	int pc = 0; // parameter counter
	if (! v.empty())
	{
		for (i=v.begin(); i<v.end(); i++)
		{
			cs = *i;
			if (left(cs,1) == (string)":")
			{
				pp = mid(cs,2);
			}
			else if (pp != (string)"")
			{
				m.insert( pair<string,string>(pp, cs) );
				pp = "";
			}
			else
			{
				m.insert( pair<string,string>(str(pc), cs) );
				pc++;
			}
		}
	}
	return m;
}

void ccp_dump_vector(const vector<string> & v)
{
	vector<string>::iterator i;
	for (i=v.begin(); i<v.end(); i++)
	{
		cout << *i << endl;
	}
}

void ccp_dump_map(map<string,string> & m)
{
	map<string,string>::iterator i;
	for (i=m.begin(); i != m.end(); i++)
	{
		cout << i->first << ": " << i->second << endl;
	}
}

