/*

 CHASM C/P - CHASM Command Parser
 Copyright (c) 2004 PXI Project Team

 Originally written in RealBasic and ported to C++ by Jonathan Bettencourt.

 This parses a string of xTalk code according to command definitions in what's called
 a "vocabulary file," whose format was written by Jonathan Bettencourt.

 To turn a line of xTalk into a vector, either:
 * Call ccp_init(), then pass a script line to ccp_parse(<script line>, <parsing mode>).
 * Create an instance of chasm_vocabulary, load a vocabulary file with the readfile(<file name>)
   method, and parse with chasm_parse(<script line>, <parsing mode>).

 The parsing mode is an optional parameter. There are three possible values:

     0 (or the constant CCP_PARSE_NUMERIC) - Puts parameters in a numbered order.
     If a parameter has no number, it appears after the numbered parameters as a set of two
     strings, one with a colon and the parameter name, and the second with the value.

     1 (or the constant CCP_PARSE_NOMINATIVE) - Assigns names to all parameters,
     regardless of whether they have numbers or not. Each parameter is defined by a set of
     two strings, one with a colon and the parameter name, and the second with the value.

     2 (or the constant CCP_PARSE_DUAL) - Combines the first two modes. All numbered
     parameters will be put in order, followed by all parameters with their names,
     regardless of whether they have numbers or not.

 Permission is hereby granted, free of charge, to any person obtaining a copy of
 this software and associated documentation files (the "Software"), to deal in the
 Software without restriction, including without limitation the rights to use,
 copy, modify, merge, publish, distribute, sublicense, and/or sell copies of the
 Software, and to permit persons to whom the Software is furnished to do so,
 subject to the following conditions:

 The above copyright notice and this permission notice shall be included in all copies
 or substantial portions of the Software.

 THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED,
 INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A
 PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT
 HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN ACTION
 OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN CONNECTION WITH THE
 SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#ifndef _____CHASM_CP_____
#define _____CHASM_CP_____

#include <string>
#include <vector>
#include <map>
#include <stack>
#include "stringmaniplib.h"
#include "binarystream.h"
#include "bchasm.h"
#include "variant.h"
using namespace std;

const int CCP_PARSE_NUMERIC = 0;
const int CCP_PARSE_NOMINATIVE = 1;
const int CCP_PARSE_BOTH = 2;
const int CCP_PARSE_DUAL = 2;

/* CHASM Vocabulary Word
   This is a class that holds all the information about how
   to parse a certain command. */

class chasm_vocab_word
{
	public:

	string name;
	int paramcount;
	vector<string> params;
	string type;

	string gparams(int);
	string htequiv(void);
	string kind(void);
	int sortpriority(void);
	void readdata(const string &);
	void sparams(int, string);
	string writedata(void);
};

/* CHASM Vocabulary
   This is the class that actually parses commands, using
   vocabulary defined by chasm_vocab_words. */

class chasm_vocabulary
{
	public:

	bool convertoptionl;
	bool stripSpaces;
	bool processAssignment;
	bool processFunctiOn;
	vector<chasm_vocab_word> wordz;

	chasm_vocabulary(void);
	chasm_vocabulary(const string &);

	vector<string> chasm_parse(const string &);
	vector<string> chasm_parse(const string &, int);
	void clear(void);
	string indextoname(int);
	int namekindtoindex(const string &, const string &);
	int nametoindex(const string &);
	int nametypetoindex(const string &, const string &);
	void readdata(const string &);
	void readfile(const string &);
	void sort(void);
	void sortbyname(void);
	string writedata(void);
	void writefile(const string &);

	const vector<chasm_vocab_word>::iterator begin(void);
	const vector<chasm_vocab_word>::iterator end(void);

	private:

	string findFactor(const string &);
	string findFactor(const string &, const string &);
	int findFactor(const vector<string> &, int);
	vector<string> ccp_jrrTolkienizer(const string &);
	string ccp_jrrDeTolkienizer(const vector<string> &);

	int countchars(const string &, const string &);
	bool equalparens(const string &);
	bool equalquotes(const string &);
	bool balancedexpr(const string &);
	string stripdupspaces(const string &);
	int mbwinstr(const string &, const string &);
	bool mbwsatisfied(const string &, const string &);

	union hasher
	{
		char char_val[4];
		long int hash_val;
	};
};

void ccp_init(void);
vector<string> ccp_parse(const string &);
vector<string> ccp_parse(const string &, int);
map<string,string> ccp_vector_to_map(const vector<string> &);
void ccp_dump_vector(const vector<string> &);
void ccp_dump_map(map<string,string> &);

#endif